'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'   MODEXPCX.BAS -- created in QBasic 1.1 by Steven Sensarn
'      * speed has been slightly compromised for size
'      * additional Mode X demos from MODEX.BAS and FADE.BAS added.
'
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

DEFINT A-Z

'//STRUCTURES

TYPE THeader            '= signs are for loading requirements
    MAN AS STRING * 1   'manufacturer = 10
    VER AS STRING * 1   'version = 5+
    ENC AS STRING * 1   'encoding = 1 (RLE)
    BPP AS STRING * 1   'bits per pixel per plane = 8
    MNX AS INTEGER      'minimum X size
    MNY AS INTEGER      'minimum Y size
    MAX AS INTEGER      'maximum X size
    MAY AS INTEGER      'maximum Y size
    HRS AS INTEGER      'horizontal resolution
    VRS AS INTEGER      'vertical resolution
    COL AS STRING * 48  'EGA color palette (ignored)
    RES AS STRING * 1   'reserved
    NMP AS STRING * 1   'number of planes = 1
    BPL AS INTEGER      'bytes per scanline
    PLI AS INTEGER      'palette info (ignored)
    FIL AS STRING * 58  'filler
END TYPE

TYPE PaletteType
      Red AS INTEGER
      Green AS INTEGER
      Blue AS INTEGER
END TYPE

'//VARIABLE DECLARATIONS
DIM Byte AS STRING * 1    'used to read bytes
DIM Header AS THeader     'used for header info
DIM index AS LONG         'used to control decoding
DIM size AS LONG          'used to control decoding
DIM XSize AS INTEGER, YSize AS INTEGER 'image size
DIM totalbytes AS INTEGER 'used to measure scan lines
DIM value AS INTEGER      'used for converting char to int
DIM rlp                   'used for run length loop
DIM X AS LONG, Y AS LONG  'used for pixel tracking
DIM SHARED pData(0 TO 255, 1 TO 3)
DIM SHARED Pal AS PaletteType
DIM SHARED FILE AS STRING

'//FUNCTION PROTOTYPES
DECLARE SUB X.PLANE (p AS INTEGER)        'use this to change planes
DECLARE SUB X.PIXEL (X AS LONG, Y AS LONG, CLR AS INTEGER)  'plots pixel
DECLARE SUB MODEX ()                      'changes to Mode X 320x240x256
DECLARE SUB WAIT.KEY ()                   'press any key to continue
DECLARE SUB OPEN.FILE ()                  'used to load PCX file
DECLARE SUB SetVOffset (Offset&)          'Mode X smooth vertical scrolling
DECLARE SUB OutPort (PortAddr&, vh%, vl%) 'writes to VGA ports
DECLARE SUB Palette.Set (nColor%, pInfo AS PaletteType)
DECLARE SUB Palette.Get (nColor%, pInfo AS PaletteType)
DECLARE SUB Palette.Fadeout ()
DECLARE SUB Palette.FadeIn ()
DECLARE SUB PCX.INFO (Header AS THeader)

'//CONSTANTS
CONST OffsetInc% = 80           'used in vertical scrolling

'//VARIABLE INITIALIZATIONS
index = 1                       'be on the SAFE side
FILE = "MCLAREN.PCX"




'//MAIN ==============================================================
'ON ERROR GOTO ERROR.HANDLER    'just in case...

OPEN.FILE         'open PCX file
GET #1, 1, Header 'load header

'//BASIC TEST (NOT fool-proof) -- REQUIREMENTS
IF ASC(Header.MAN) <> 10 THEN CLOSE #1: END 'not PCX
IF ASC(Header.VER) < 5 THEN CLOSE #1: END   'not 256 color
IF ASC(Header.ENC) <> 1 THEN CLOSE #1: END  'not RLE
IF ASC(Header.NMP) <> 1 THEN CLOSE #1: END  'not supportive of Mode X planes

'//LOAD WHILE LOOP INFORMATION
XSize = Header.MAX - Header.MNX + 1 'load horizontal graphic size
YSize = Header.MAY - Header.MNY + 1 'load vertical graphic size
totalbytes = Header.BPL             'used to track end of scan line
size = 1& * XSize * YSize           'load size -- force long
IF size > 76800 THEN CLOSE : END    'graphic too large -- won't fit on screen

PCX.INFO Header

MODEX            'switch to Mode X -- 320x240x256!
DEF SEG = &HA000 'video memory access

'//CHANGE PALETTE
GET #1, LOF(1) - 768, Byte:         'load palette info
IF ASC(Byte) <> 12 THEN CLOSE : END 'No 256 color palette

FOR index = 0 TO 255
  OUT &H3C6, &HFF  'initialize VGA palette
  OUT &H3C8, index 'select color index
                   '/4 because we only need top 6 bits
  GET #1, , Byte: OUT &H3C9, INT(ASC(Byte) / 4) 'RED
  GET #1, , Byte: OUT &H3C9, INT(ASC(Byte) / 4) 'GREEN
  GET #1, , Byte: OUT &H3C9, INT(ASC(Byte) / 4) 'BLUE
NEXT index
GET #1, 1, Header 'LOAD HEADER


'//DECODE AND DISPLAY
WHILE index <= size                'While the picture isn't done loading...
    GET #1, , Byte                 'read a byte from the PCX file
    IF (ASC(Byte) AND &HC0) = &HC0 THEN 'test top two bits for 1'S (RLE?)
        rlp = ASC(Byte) AND &H3F           'set run length
        GET #1, , Byte                     'read data byte
        WHILE rlp > 0                'while run length has not yet finished
            IF NOT X > XSize THEN CALL X.PIXEL(X, Y, ASC(Byte)) 'plot pixel
            X = X + 1                                'increment pixel position
            IF X >= totalbytes THEN X = 0: Y = Y + 1 'test for end of scanline
            rlp = rlp - 1                            'one less pixel to plot in run
            index = index + 1                        'increase total
        WEND
    ELSE             'if the byte is not encoded as RLE (then it's easy!)
        IF NOT X > XSize THEN CALL X.PIXEL(X, Y, ASC(Byte))  'plot one pixel
        X = X + 1                                  'renew pixel position
        IF X >= totalbytes THEN X = 0: Y = Y + 1   'test for end of scanline
        index = index + 1                          'increase total
    END IF
WEND
'scroll up then down (Isn't it smooth!?)
FOR lp = 1 TO 256
  Offset& = Offset& + OffsetInc%
  CALL SetVOffset(Offset&): WAIT &H3DA, 8
NEXT
FOR lp = 1 TO 256
  Offset& = Offset& - OffsetInc%
  CALL SetVOffset(Offset&): WAIT &H3DA, 8
NEXT
'return to normal
CALL SetVOffset(0)
SLEEP 3
Palette.Fadeout
CLOSE #1
SCREEN 0: WIDTH 80
END

'//ERRORS
ERROR.HANDLER:
SCREEN 0: WIDTH 80: CLS
PRINT
PRINT "Something has occured in the program that has"
PRINT "caused it to halt.  Terminating..."
WAIT.KEY
END


'//NOTICE
'I HAVE ABSOLUTELY NO EXPERIENCE WITH PLANE BASED PCX FILES BECAUSE I USE
'MICROSOFT WINDOWS 3.1 PAINTBRUSH AND I CAN'T SEEM TO CREATE ONE.  IF
'ANYBODY WOULD BE WILLING TO DRAW ME ONE THAT USES PLANES, I WOULD GREATLY
'APPRECIATE IT.  SEND TO "txs53132@bayou.uh.edu" VIA BINHEX OR MIME.

'Speed optimization and code cleanup by Toshi
'     * the 2^P was replaced by a select case statement
'       (couldn't get DIM powersof2(4) method to work)
'     * palette and scrolling procedures were added
'     * comments and capitalization changed for elegance
'     * PCX.INFO was put in a subroutine
'  --- Toshihiro Horie 12/8/96 (toshiman@uclink4.berkeley.edu)

DEFSNG A-Z
SUB MODEX
    SCREEN 13
    DEF SEG = &HA000
    OUT &H3C4, &H4
    OUT &H3C5, &H6
    OUT &H3C4, &H2
    OUT &H3C5, &HF
    CLS
    OUT &H3D4, &H14: OUT &H3D5, &H0
    OUT &H3D4, &H17: OUT &H3D5, &HE3
    OUT &H3C2, &HE3
    OUT &H3D4, &H11: OUT &H3D5, &H2C
    OUT &H3D4, &H6: OUT &H3D5, &HD
    OUT &H3D4, &H7: OUT &H3D5, &H3E
    OUT &H3D4, &H10: OUT &H3D5, &HEA
    OUT &H3D4, &H11: OUT &H3D5, &HAC
    OUT &H3D4, &H12: OUT &H3D5, &HDF
    OUT &H3D4, &H15: OUT &H3D5, &HE7
    OUT &H3D4, &H16: OUT &H3D5, &H6
END SUB

SUB OPEN.FILE
CLS
    PRINT "<>/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\<>"
    PRINT "<> PLEASE NOTE THAT THE FILE THAT IS TO BE LOADED MUST HAVE A   <>"
    PRINT "<> SIZE NO LARGER THAN 320x240.  IF YOU WOULD LIKE TO TERMINATE <>"
    PRINT "<> THIS PROGRAM, DO SO ENTERING AN EMPTY FILENAME.              <>"
    PRINT "<>\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/<>"
    PRINT
    FILE = RTRIM$(FILE)
    FILE = UCASE$(FILE)
    IF NOT RIGHT$(RTRIM$(FILE), 4) = ".PCX" THEN END 'no valid extension
    OPEN FILE FOR BINARY AS #1
    IF LOF(1) = 0 THEN CLOSE #1: KILL FILE: END 'FILE EMPTY OR NOT FOUND

END SUB

DEFINT A-Z
SUB OutPort (PortAddr&, vh%, vl%)
  OUT PortAddr&, vl%: OUT PortAddr& + 1, vh%
END SUB

SUB Palette.FadeIn
DIM tT(1 TO 3)
FOR I = 1 TO 64
WAIT &H3DA, 8, 8
  FOR O = 0 TO 255
    Palette.Get O, Pal
    tT(1) = Pal.Red
    tT(2) = Pal.Green
    tT(3) = Pal.Blue
    IF tT(1) < pData(O, 1) THEN tT(1) = tT(1) + 1
    IF tT(2) < pData(O, 2) THEN tT(2) = tT(2) + 1
    IF tT(3) < pData(O, 3) THEN tT(3) = tT(3) + 1
    Pal.Red = tT(1)
    Pal.Green = tT(2)
    Pal.Blue = tT(3)
    Palette.Set O, Pal
  NEXT O
NEXT I

END SUB

SUB Palette.Fadeout
DIM tT(1 TO 3)
FOR I = 0 TO 255
  Palette.Get I, Pal
  pData(I, 1) = Pal.Red
  pData(I, 2) = Pal.Green
  pData(I, 3) = Pal.Blue
NEXT I
FOR I = 1 TO 64
WAIT &H3DA, 8, 8
  FOR O = 0 TO 255
    Palette.Get O, Pal
    tT(1) = Pal.Red
    tT(2) = Pal.Green
    tT(3) = Pal.Blue
    IF tT(1) > 0 THEN tT(1) = tT(1) - 1
    IF tT(2) > 0 THEN tT(2) = tT(2) - 1
    IF tT(3) > 0 THEN tT(3) = tT(3) - 1
    Pal.Red = tT(1)
    Pal.Green = tT(2)
    Pal.Blue = tT(3)
    Palette.Set O, Pal
  NEXT O
NEXT I
END SUB

SUB Palette.Get (nColor%, pInfo AS PaletteType)
OUT &H3C6, &HFF
OUT &H3C7, nColor%
pInfo.Red = INP(&H3C9)
pInfo.Green = INP(&H3C9)
pInfo.Blue = INP(&H3C9)

END SUB

SUB Palette.Set (nColor%, pInfo AS PaletteType)
OUT &H3C6, &HFF
OUT &H3C8, nColor%
OUT &H3C9, pInfo.Red
OUT &H3C9, pInfo.Green
OUT &H3C9, pInfo.Blue
END SUB

SUB PCX.INFO (Header AS THeader)
'//display PCX file info
CLS : SCREEN 0: WIDTH 80
COLOR 14, 4: PRINT "MODE X PCX LOADER by Steven Sensarn       "
COLOR 15, 1
PRINT "FILENAME: ", , " "; FILE: PRINT
COLOR 7, 0
PRINT "MANUFACTURER: ", ASC(Header.MAN)
PRINT "VERSION: ", , ASC(Header.VER)
PRINT "RLE ENCODING: ", ASC(Header.ENC)
PRINT "BITS PER PIXEL PER PLANE: ", ASC(Header.BPP)
PRINT "WINDOW X MIN: ", Header.MNX
PRINT "WINDOW Y MIN: ", Header.MNY
PRINT "WINDOW X MAX: ", Header.MAX
PRINT "WINDOW Y MAX: ", Header.MAY
PRINT "HORIZONTAL RESOLUTION: ", Header.HRS
PRINT "VERTICAL RESOLUTION: ", Header.VRS
PRINT "NUMBER OF PLANES: ", ASC(Header.NMP)
PRINT "BYTES PER LINE: ", Header.BPL
PRINT
IF NOT Header.HRS / 320 = INT(Header.HRS / 320) THEN
    'WARPED IMAGE -- NO BIG DEAL
    PRINT "ASPECT RATIO MISMATCH ", " -- X AXIS"
ELSE
    PRINT
END IF
IF NOT Header.VRS / 240 = INT(Header.VRS / 240) THEN
    'WARPED IMAGE -- NO BIG DEAL
    PRINT "ASPECT RATIO MISMATCH ", " -- Y AXIS"
END IF
SLEEP 1
'WAIT.KEY

END SUB

SUB SetVOffset (Offset&)
  V& = Offset&
  CALL OutPort(&H3D4, V& \ 256, 12)
  CALL OutPort(&H3D4, V& AND 255, 13)
END SUB

DEFSNG A-Z
SUB WAIT.KEY
  WHILE LEN(INKEY$) = 0: WEND
END SUB

'//ASSUMING CURRENT SEGMENT IS A000:0000H
SUB X.PIXEL (X AS LONG, Y AS LONG, CLR AS INTEGER)
    CALL X.PLANE(X AND 3)
    POKE INT(Y * 80 + X / 4), CLR
END SUB

'//USED TO CHANGE PLANE
SUB X.PLANE (p AS INTEGER)
OUT &H3C4, 2
SELECT CASE p
      CASE 0
            OUT &H3C5, 1
      CASE 1
            OUT &H3C5, 2
      CASE 2
            OUT &H3C5, 4
      CASE 3
            OUT &H3C5, 8
END SELECT
END SUB

